"""
Help generate various Update objects used in testing
"""
import copy, datetime

from meetg.api_types import (
    ApiType, ChatApiType, MessageApiType, PhotoSizeApiType, UpdateApiType, UserApiType,
)
from meetg.loging import get_logger
from meetg.utils import parse_entities


logger = get_logger()

LAST_INT = 0


def get_next_int():
    """Always return a larger int. Useful to generate real-looking IDs"""
    global LAST_INT
    LAST_INT += 1
    return LAST_INT


class Factory:
    """
    Base class for any factory of a PTB object
    """
    def __init__(self, tgbot):
        self.tgbot = tgbot

    def _filter_prefix(self, data, prefix):
        filtered = {}
        length = len(prefix)
        for key, val in data.items():
            if key.startswith(prefix):
                filtered[key[length:]] = val
        return filtered

    def _create_args(self, kwargs):
        validated_data = self.api_type(kwargs).validated_data
        args = self.get_defaults()
        args.update(validated_data)
        return args

    def create(self, **kwargs):
        args = self._create_args(kwargs)
        obj = self.api_type.ptb_class(**args)
        return obj


class UpdateFactory(Factory):
    """
    Base factory of Update objects
    """
    api_type = UpdateApiType
    update_type = None


class ChatFactory(Factory):
    api_type = ChatApiType

    def get_defaults(self):
        defaults = {
            'id': get_next_int(),
            'type': 'private',
        }
        return defaults


class UserFactory(Factory):
    api_type = UserApiType

    def get_defaults(self):
        defaults = {
            'id': get_next_int(),
            'first_name': 'Palin',
            'is_bot': False,
        }
        return defaults


class PhotoSizeFactory(Factory):
    api_type = PhotoSizeApiType

    def get_defaults(self):
        defaults = {
            'file_id': generate_random_string(98),
            'file_unique_id': generate_random_string(19),
            'width': 1,
            'height': 1,
            'file_size': 1,
        }
        return defaults


class MessageFactory(Factory):
    api_type = MessageApiType

    def __init__(self, tgbot, message_type):
        super().__init__(tgbot)
        self.message_type = message_type

    def get_defaults(self):
        defaults = {
            'message_id': get_next_int(),
            'text': 'Spam',
            'date': datetime.datetime.now(),
            'bot': self.tgbot,
        }
        if self.message_type == 'edited_message':
            defaults['edit_date'] = datetime.datetime.now()
        return defaults

    def create(self, **kwargs):
        chat_args = self._filter_prefix(kwargs, 'chat__')
        photo_size_args = self._filter_prefix(kwargs, 'photo_size__')
        from_user_args = self._filter_prefix(kwargs, 'from__')
        if 'id' in chat_args and chat_args['id'] > 0 and 'id' not in from_user_args:
            from_user_args['id'] = chat_args['id']

        args = self._create_args(kwargs)
        args['chat'] = ChatFactory(self.tgbot).create(**chat_args)
        args['from_user'] = UserFactory(self.tgbot).create(**from_user_args)
        args['entities'] = parse_entities(args['text'])
        if photo_size_args:
            args['photo'] = [PhotoSizeFactory(self.tgbot).create(**photo_size_args)]

        obj = self.api_type.ptb_class(**args)
        return obj


class MessageUpdateFactory(UpdateFactory):
    """
    Factory of an Update with a field of type 'Message'.
    It is a: message, edited_message, channel_post, and edited_channel_post
    """
    api_type = UpdateApiType

    update_type = 'Message'
    type_fields = MessageApiType.fields

    def __init__(self, tgbot, message_type):
        super().__init__(tgbot)
        self.message_type = message_type

    def create(self, **kwargs):
        message = MessageFactory(self.tgbot, self.message_type).create(**kwargs)
        update_obj = self.api_type.ptb_class(get_next_int(), **{self.message_type: message})
        return update_obj


class InlineQueryUpdateFactory(UpdateFactory):
    """
    Factory of an Update with a field of type 'InlineQuery'
    """
    update_type = 'InlineQuery'


class ChosenInlineResultUpdateFactory(UpdateFactory):
    """
    Factory of an Update with a field of type 'ChosenInlineResult'
    """
    update_type = 'ChosenInlineResult'



import time

import telegram

import settings
from meetg.loging import get_logger


logger = get_logger()


class ApiMethod:

    def __init__(self, tgbot, is_mock=False):
        self.tgbot = tgbot
        self.is_mock = is_mock
        self.args = None

    def easy_call(self, **kwargs):
        """Call the method by simplified params"""
        raise NotImplementedError

    def log(self, kwargs):
        raise NotImplementedError

    def call(self, **kwargs):
        """
        Call the method by the exact Telegram API params,
        by keyword arguments only, to easily validate them
        """
        self.args = kwargs.copy()
        kwargs = self._validate(kwargs)
        if self.is_mock:
            success, response = None, None
        else:
            success, response = self._call(kwargs)
        if success:
            self.log(self.args)
        return success, response

    def __str__(self):
        if self.args:
            return f'{self.name}: {self.args}'
        else:
            return f'{self.name}: no args'

    def _call(self, kwargs):
        """
        Retries, handling network and load issues
        """
        to_attempt = settings.api_attempts
        success, response = False, None
        tgbot_method = getattr(self.tgbot, self.name)

        while to_attempt > 0:
            try:
                response = tgbot_method(**kwargs)
                success = True
                to_attempt = 0
            except telegram.error.NetworkError as exc:
                prefix = 'Network error: '
                if 'are exactly the same as' in exc.message:
                    logger.error(prefix + '"%s". It\'s ok, nothing to do here', exc.message)
                    success = True
                    to_attempt = 0
                elif "Can't parse entities" in exc.message:
                    logger.error(prefix + '"%s". Retrying is pointless', exc.message)
                    to_attempt = 0
                elif "Message to forward not found" in exc.message:
                    logger.error(prefix + '"%s". Retrying is pointless', exc.message)
                    to_attempt = 0
                else:
                    logger.error(
                        prefix + '"%s". Waiting %s seconds then retry',
                        settings.network_error_wait, exc.message,
                    )
                    to_attempt -= 1
                    time.sleep(settings.network_error_wait)
                response = exc.message
            except telegram.error.TimedOut as exc:
                logger.error('Timed Out. Retrying')
                response = exc.message
                to_attempt -= 1
            except telegram.error.RetryAfter as exc:
                logger.error('It is asked to retry after %s seconds. Doing', exc.retry_after)
                response = exc.message
                to_attempt -= 2
                time.sleep(exc.retry_after + 1)
            except telegram.error.ChatMigrated as exc:
                logger.error('ChatMigrated error: "%s". Retrying with new chat id', exc)
                response = exc.message
                kwargs['chat_id'] = exc.new_chat_id
                to_attempt -= 1
            except (telegram.error.Unauthorized, telegram.error.BadRequest) as exc:
                logger.error('Error: "%s". Retrying', exc)
                response = exc.message
                to_attempt -= 2
        logger.debug('Success' if success else 'Fail')
        return success, response

    def _validate(self, data):
        validated = {}
        for key in data:
            if key in self.parameters:
                validated[key] = data[key]
            else:
                logger.warning('Method %s doesn\'t accept arg %s', self.name, key)
        return validated


class SendMessageMethod(ApiMethod):
    name = 'send_message'
    parameters = (
        # required
        'chat_id', 'text',
        # optional
        'parse_mode', 'entities', 'disable_web_page_preview', 'disable_notification',
        'reply_to_message_id', 'allow_sending_without_reply', 'reply_markup', 
    )

    def easy_call(self, chat_id, text, reply_to=None, markup=None, html=None, preview=False):
        parse_mode = telegram.ParseMode.HTML if html else None
        success, response = self.call(
            chat_id=chat_id, text=text, reply_to_message_id=reply_to, reply_markup=markup,
            parse_mode=parse_mode, disable_web_page_preview=not preview,
        )
        return success, response

    def log(self, kwargs):
        chat_id = kwargs.get('chat_id')
        text = repr(kwargs.get('text', ''))
        logger.info('Send message to chat %s, text length %s', chat_id, len(text))


class EditMessageTextMethod(ApiMethod):
    name ='edit_message_text'
    parameters = (
        # required
        'text',
        # optional
        'chat_id', 'message_id', 'inline_message_id', 'parse_mode', 'entities',
        'disable_web_page_preview', 'reply_markup',
    )
    def easy_call(self, text, chat_id, message_id, preview=False):
        success, response = self.call(
            text=text, chat_id=chat_id, message_id=message_id,
            disable_web_page_preview=not preview,
        )
        return success, response

    def log(self, kwargs):
        chat_id = kwargs.get('chat_id')
        message_id = kwargs.get('message_id')
        logger.info('Edit message %s in chat %s', message_id, chat_id)


class DeleteMessageMethod(ApiMethod):
    name = 'delete_message'
    parameters = (
        # required
        'chat_id', 'message_id',
    )
    def easy_call(self, chat_id, message_id):
        success, response = self.call(
            chat_id=chat_id, message_id=message_id,
        )
        return success, response

    def log(self, kwargs):
        chat_id = kwargs.get('chat_id')
        message_id = kwargs.get('message_id')
        logger.info('Delete message %s in chat %s', message_id, chat_id)


class ForwardMessageMethod(ApiMethod):
    name = 'forward_message'
    parameters = (
        # required
        'chat_id', 'from_chat_id', 'message_id',
        # optional
        'disable_notification',
    )
    def easy_call(self, chat_id, from_chat_id, message_id, notify=True):
        success, response = self.call(
            chat_id=chat_id, from_chat_id=from_chat_id, message_id=message_id,
            disable_notification=not notify,
        )
        return success, response

    def log(self, kwargs):
        chat_id = kwargs.get('chat_id')
        from_chat_id = kwargs.get('from_chat_id')
        message_id = kwargs.get('message_id')
        logger.info(
            'Forward message %s from chat %s to chat %s',
            message_id, from_chat_id, chat_id,
        )


class SendPhotoMethod(ApiMethod):
    name = 'send_photo'
    parameters = (
        # required
        'chat_id', 'photo',
        # optional
        'caption', 'parse_mode', 'caption_entities', 'disable_notification', 'reply_to_message_id',
        'allow_sending_without_reply', 'reply_markup',
    )

    def easy_call(self, chat_id, photo, caption=None, reply_to=None, markup=None, html=None):
        parse_mode = telegram.ParseMode.HTML if html else None
        success, response = self.call(
            chat_id=chat_id, photo=photo, caption=caption, reply_to_message_id=reply_to,
            reply_markup=markup, parse_mode=parse_mode,
        )
        return success, response

    def log(self, kwargs):
        chat_id = kwargs.get('chat_id')
        logger.info('Send photo to chat %s', chat_id)


class SendDocumentMethod(ApiMethod):
    name = 'send_document'
    parameters = (
        # required
        'chat_id', 'document',
        # optional
        'thumb', 'caption', 'parse_mode', 'caption_entities', 'disable_content_type_detection',
        'disable_notification', 'reply_to_message_id', 'allow_sending_without_reply',
        'reply_markup',
    )

    def easy_call(
            self, chat_id, document, thumb=None, caption=None, reply_to=None, markup=None,
            html=None,
        ):
        parse_mode = telegram.ParseMode.HTML if html else None
        success, response = self.call(
            chat_id=chat_id, document=document, thumb=thumb, caption=caption,
            reply_to_message_id=reply_to, reply_markup=markup, parse_mode=parse_mode,
        )
        return success, response

    def log(self, kwargs):
        chat_id = kwargs.get('chat_id')
        logger.info('Send document to chat %s', chat_id)


api_method_classes = {
    'send_message': SendMessageMethod,
    'send_photo': SendPhotoMethod,
    'send_document': SendDocumentMethod,
    'edit_message_text': EditMessageTextMethod,
    'delete_message': DeleteMessageMethod,
    'forward_message': ForwardMessageMethod,
}



from meetg.utils import get_current_unixtime, get_unixtime_before_now


class SoftDict(dict):
    """
    Dict that doesn't raise exceptions. It returns another SoftDict
    in any usually-exception case, or custom class if init() called
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_class = SoftDict

    def __getitem__(self, key):
        if self.__contains__(key):
            item = super().__getitem__(key)
        else:
            item = self.default_class()
            self.__setitem__(key, item)
        return item

    def __delitem__(self, key):
        if self.__contains__(key):
            return super().__delitem__(key)

    def init(self, cls):
        self.default_class = cls


"""App-wide cache"""
service_cache = SoftDict()


class DateSegment(list):
    """
    List with two dates: one on object creation,
    one on calling the finish()
    """
    def __init__(self):
        super().__init__()
        time = get_current_unixtime()
        self.append(time)

    def finish(self):
        time = get_current_unixtime()
        self.append(time)

    def get_duration(self):
        return self[-1] - self[0]


class DateCache(list):
    """
    List to easy store and get back date objects: points and segments.
    Useful to store runtime stats, to later report them.
    Usage:
    cache = DateCache()
    cache.add() - add current Unix time to cache
    cache.get_day_count() - return number of additions happened for the last 24 hours
    """
    def add(self, obj=None):
        if obj is None:
            obj = get_current_unixtime()
        return self.append(obj)

    def _get_day_treshold(self):
        day_before = get_unixtime_before_now(24)
        for i, item in enumerate(self):
            to_compare = item
            if isinstance(item, list):
                to_compare = item[-1]
            if to_compare >= day_before:
                return i

    def get_day(self):
        treshold = self._get_day_treshold()
        if treshold is not None:
            return self[treshold:]

    def get_day_count(self):
        last_day = self.get_day()
        return len(last_day)

    def get_day_duration(self):
        """
        Method only for Segments. Count duration
        of date segments added for the last 24 hours
        """
        last_day = self.get_day()
        total = 0
        for segment in last_day:
            if not isinstance(segment, DateSegment):
                raise RuntimeError('get_day_duration() is only applicable for DateSegments')
            total += segment.get_duration()
        return total

    def clear_before_last_day(self):
        treshold = self._get_day_treshold()
        if treshold is not None:
            del self[:treshold]


def get_job_reports():
    """Get gathered info from service_cache['stats']['job'] and format it"""
    reports = []
    for job_name, segments in service_cache['stats']['job'].items():
        segments.clear_before_last_day()
        duration = segments.get_day_duration()
        line = f'{job_name} took {duration:.3f} seconds total'
        reports.append(line)
    return reports


def get_update_reports():
    """Get gathered info from service_cache['stats']['update'] and format it"""
    update_reports = []
    for update_type, dates in service_cache['stats']['update'].items():
        dates.clear_before_last_day()
        count = dates.get_day_count()
        line = f"received {count} '{update_type}' updates"
        update_reports.append(line)
    return update_reports


class _SaveTimeJobQueueWrapper:
    """
    A wrapper to measure job time execution,
    to report it in stats
    """
    def __init__(self, job_queue):
        self.job_queue = job_queue
        self._wrapped_callbacks = []

    def _wrap(self, callback, *args, **kwargs):

        def wrapped(*args, **kwargs):
            service_cache['stats']['job'].init(DateCache)
            segment = DateSegment()
            result = callback(*args, **kwargs)
            segment.finish()
            service_cache['stats']['job'][callback.__name__].add(segment)
            return result

        wrapped.__doc__ = callback.__doc__
        wrapped.__name__ = callback.__name__
        wrapped.__module__ = callback.__module__
        wrapped.__qualname__ = callback.__qualname__
        wrapped.__annotations__ = callback.__annotations__
        self._wrapped_callbacks.append(wrapped)
        return wrapped

    def run_once(self, callback, *args, **kwargs):
        wrapped = self._wrap(callback, *args, **kwargs)
        return self.job_queue.run_once(wrapped, *args, **kwargs)

    def run_repeating(self, callback, *args, **kwargs):
        wrapped = self._wrap(callback)
        return self.job_queue.run_repeating(wrapped, *args, **kwargs)

    def run_monthly(self, callback, *args, **kwargs):
        wrapped = self._wrap(callback)
        return self.job_queue.run_monthly(wrapped, *args, **kwargs)

    def run_daily(self, callback, *args, **kwargs):
        wrapped = self._wrap(callback)
        return self.job_queue.run_daily(wrapped, *args, **kwargs)

    def run_custom(self, callback, *args, **kwargs):
        wrapped = self._wrap(callback)
        return self.job_queue.run_custom(wrapped, *args, **kwargs)



import datetime, importlib, logging, unittest

from telegram import Chat, Message, Update, User

import settings
from meetg import default_settings
from meetg.loging import get_logger
from meetg.storage import get_model_classes
from meetg.utils import dict_to_obj, import_string, parse_entities


class BaseTestCase(unittest.TestCase):

    def _reset_settings(self):
        importlib.reload(settings)

    def _drop_db(self):
        for model_class in get_model_classes():
            Model = import_string(model_class)
            Model(test=True).drop()

    def _reinit_loggers(self):
        import meetg.botting
        import meetg.storage
        logger = get_logger()
        meetg.botting.logger = logger
        meetg.storage.logger = logger

    def setUp(self):
        super().setUp()
        self._reset_settings()
        settings.log_level = logging.ERROR
        self._reinit_loggers()
        self._drop_db()

    def tearDown(self):
        super().tearDown()
        self._drop_db()


class BotTestCase(BaseTestCase):

    def setUp(self):
        super().setUp()
        Bot = import_string(settings.bot_class)
        self.bot = Bot(mock=True)


class JobQueueMock:

    def run_daily(self, callback, period):
        pass

    def run_repeating(self, callback, period):
        pass

    def run_monthly(self, callback, period):
        pass

    def run_once(self, callback, period):
        pass


class TgBotMock:
    """A mock for PTB Bot"""
    username = 'mock_username'

    def __getattr__(self, name):
        pass

    def get_me(self):
        me = dict_to_obj('Me', {'username': self.username})
        return me


class UpdaterMock:
    """A mock for PTB Updater"""

    def __init__(self, *args, **kwargs):
        self.job_queue = JobQueueMock()
        self.bot = TgBotMock()



"""
Universal classes describing Telegram Bot API types.
Useful in many places: storage, factories, etc
"""
from telegram import Chat, Message, PhotoSize, Update, User


class ApiType:
    """Base class for any API type class"""
    ptb_class = None

    def __init__(self, data):
        self.validated_data = self.validate(data)

    def validate(self, data):
        validated = {}
        for key in data:
            if key in self.fields:
                validated[key] = data[key]
        return validated

    def get_ptb(self):
        ptb_obj = self.ptb_class(**self.validated_data)
        return ptb_obj


class UpdateApiType(ApiType):
    name = 'Update'
    ptb_class = Update
    id_field = 'update_id'
    fields = (
        # required
        'update_id',
        # optional
        'message', 'edited_message', 'channel_post', 'edited_channel_post', 'inline_query',
        'chosen_inline_result', 'callback_query', 'shipping_query', 'pre_checkout_query', 'poll',
        'poll_answer',
    )


class MessageApiType(ApiType):
    name = 'Message'
    ptb_class = Message
    id_field = 'message_id'
    fields = (
        # required
        'message_id', 'date', 'chat',
        # optional
        'from', 'sender_chat', 'forward_from', 'forward_from_chat', 'forward_from_message_id',
        'forward_signature', 'forward_sender_name', 'forward_date', 'reply_to_message', 'via_bot',
        'edit_date', 'media_group_id', 'author_signature', 'text', 'entities', 'animation',
        'audio', 'document', 'photo', 'sticker', 'video', 'video_note', 'voice', 'caption',
        'caption_entities', 'contact', 'dice', 'game', 'poll', 'venue', 'location',
        'new_chat_members', 'left_chat_member', 'new_chat_title', 'new_chat_photo',
        'delete_chat_photo', 'group_chat_created', 'supergroup_chat_created',
        'channel_chat_created', 'migrate_to_chat_id', 'migrate_from_chat_id', 'pinned_message',
        'invoice', 'successful_payment', 'connected_website', 'passport_data',
        'proximity_alert_triggered', 'reply_markup',
    )


class UserApiType(ApiType):
    name = 'User'
    ptb_class = User
    id_field = 'id'
    fields = (
        # required
        'id', 'is_bot', 'first_name',
        # optional
        'last_name', 'username', 'language_code', 'can_join_groups', 'can_read_all_group_messages',
        'supports_inline_queries',
    )


class ChatApiType(ApiType):
    name = 'Chat'
    ptb_class = Chat
    id_field = 'id'
    fields = (
        # required
        'id', 'type',
        # optional
        'title', 'username', 'first_name', 'last_name', 'photo', 'bio', 'description',
        'invite_link', 'pinned_message', 'permissions', 'slow_mode_delay', 'sticker_set_name',
        'can_set_sticker_set', 'linked_chat_id', 'location', 'all_members_are_administrators',
    )


class PhotoSizeApiType(ApiType):
    name = 'PhotoSize'
    ptb_class = PhotoSize
    id_field = 'file_id'
    fields = (
        'file_id', 'file_unique_id', 'width', 'height', 'file_size',
    )



import random, string, time
from collections import namedtuple
from importlib import import_module
from io import BytesIO

from PIL import Image
from telegram.messageentity import MessageEntity


def generate_random_string(length=22):
    population = string.ascii_letters + string.digits
    choices = random.choices(population, k=length)
    random_string = ''.join(choices)
    return random_string


def message_contain(msg, words):
    msg_words = msg.lower().split()
    for msg_word in msg_words:
        for word in words:
            if msg_word.startswith(word):
                return True


def message_starts(msg, words):
    msg = msg.lower()
    for word in words:
        if msg.startswith(word):
            return True


def frange(start, stop=None, step=None):
    while stop > start:
        yield start
        start += step


def get_current_unixtime():
    return time.time()


def import_string(dotted_path):
    module_path, class_name = dotted_path.rsplit('.', 1)
    module = import_module(module_path)
    return getattr(module, class_name)


def dict_to_obj(name, dictionary: dict):
    return namedtuple(name, dictionary.keys())(*dictionary.values())


def parse_entities(string):
    """
    In real world, Telegram server does it.
    But for testing we may automatically create them
    to not create in tests each time.
    """
    entities = []
    if string.startswith('/'):
        entity = MessageEntity(
            type=MessageEntity.BOT_COMMAND,
            offset=0,
            length=len(string.split()[0]),
        )
        entities.append(entity)
    return entities


def get_unixtime_before_now(hours: int):
    before = time.time() - hours * 60 * 60
    return before


def get_update_type(update_obj):
    for key in update_obj.to_dict():
        if key != 'update_id':
            return key


def get_1px_image():
    file = BytesIO()
    image = Image.new('1', size=(1, 1))
    image.save(file, 'png')
    file.seek(0)
    return file



import time

import pymongo

import settings
from meetg.api_types import (
    ApiType, ChatApiType, MessageApiType, UpdateApiType, UserApiType,
)
from meetg.utils import get_unixtime_before_now, import_string
from meetg.loging import get_logger


logger = get_logger()


class AbstractStorage:
    """Any other storage must be a subclass of this class"""

    def __init__(self, db_name, table_name, host, port):
        self.db_name = db_name
        self.table_name = table_name
        self.host = host
        self.port = port

    def create(self, entry):
        raise NotImplementedError

    def update(self, query, update):
        raise NotImplementedError

    def update_one(self, query, update):
        raise NotImplementedError

    def count(self, query=None):
        raise NotImplementedError

    def find(self, query=None):
        raise NotImplementedError

    def find_one(self, query=None):
        raise NotImplementedError

    def delete(self, query):
        raise NotImplementedError

    def delete_one(self, query):
        raise NotImplementedError

    def drop(self):
        raise NotImplementedError


class MongoStorage(AbstractStorage):
    """
    Wrapper for MongoDB collection methods. It's some kind of an ORM.
    The idea is that another potential storage, e.g. PostgreStorage,
    have to implement the same methods, allowing the same args to them.
    But I'm not sure it will be handful. So methods and args may change in the future.
    """
    def __init__(self, db_name, table_name, host='localhost', port=27017):
        super().__init__(db_name, table_name, host, port)
        self.client = pymongo.MongoClient(host=host, port=port)
        self.db = getattr(self.client, db_name)
        self.table = getattr(self.db, table_name)

    def create(self, entry):
        return self.table.insert_one(entry)

    def update(self, query, new_data):
        return self.table.update_many(query, {'$set': new_data})

    def update_one(self, query, new_data):
        return self.table.update_one(query, {'$set': new_data})

    def count(self, query=None):
        return self.table.count_documents(query or {})

    def find(self, query=None):
        return self.table.find(query)

    def find_one(self, query=None):
        return self.table.find_one(query)

    def delete(self, query):
        return self.table.delete_many(query)

    def delete_one(self, query):
        return self.table.delete_one(query)

    def drop(self):
        return self.db.drop_collection(self.table_name)


class BaseDefaultModel:
    """Base class for default models"""

    def __init__(self, test=False):
        db_name = settings.db_name_test if test else settings.db_name
        Storage = import_string(settings.storage_class)
        self._storage = Storage(
            db_name=db_name, table_name=self.table_name,
            host=settings.db_host, port=settings.db_port,
        )

    @property
    def name_lower(self):
        return self.name.lower()

    @property
    def table_name(self):
        return f'{self.name_lower}_table'

    def _validate(self, data):
        validated = {}
        for field in data:
            if field in self.fields:
                validated[field] = data[field]
            else:
                logger.warning('Field %s doesn\'t belong to model %s', field, self.name)
        return validated

    def drop(self):
        result = self._storage.drop()
        return result

    def _log_create(self, data: dict):
        logger.info('Storage: %s created', self.name)

    def _log_update(self, data: dict):
        logger.info('Storage: %s updated', self.name)

    def create(self, data: dict):
        data = self._validate(data)
        result = None
        if data:
            data['meetg_created_at'] = time.time()
            data['meetg_modified_at'] = None
            result = self._storage.create(data)
            self._log_create(data)
        return result

    def find(self, query=None):
        found = self._storage.find(query)
        return found

    def find_one(self, query=None):
        found = self._storage.find_one(query)
        return found

    def update(self, query, new_data):
        new_data = self._validate(new_data)
        new_data['meetg_modified_at'] = time.time()
        updated = self._storage.update(query, new_data)
        return updated

    def update_one(self, query, new_data):
        new_data = self._validate(new_data)
        new_data['meetg_modified_at'] = time.time()
        updated = self._storage.update_one(query, new_data)
        self._log_update(query)
        return updated

    def count(self, query=None):
        counted = self._storage.count(query)
        return counted

    def _get_created_for_day_query(self):
        query = {
            'meetg_created_at': {
                '$lt': time.time(),
                '$gte': get_unixtime_before_now(24),
            },
        }
        return query

    def get_day_report(self):
        query = self._get_created_for_day_query()
        count = self.count(query)
        return f'stored {count} new {self.name_lower}s'


class ApiTypeModel(BaseDefaultModel):
    """Base model class for objects related to Bot API"""

    def _log_create(self, data: dict):
        id_field = self.api_type.id_field
        if id_field in data:
            logger.info('Storage: %s %s created', self.name, data[id_field])
        else:
            logger.info('Storage: %s created', self.name)

    def _log_update(self, data: dict):
        id_field = self.api_type.id_field
        if id_field in data:
            logger.info('Storage: %s %s updated', self.name, data[id_field])
        else:
            logger.info('Storage: %s updated', self.name)

    def get_ptb_obj(self, update_obj):
        raise NotImplementedError

    def get_query(self, ptb_obj):
        raise NotImplementedError

    def is_equal(self, ptb_obj, db_obj):
        equal = True
        for key, val in ptb_obj.to_dict().items():
            if key in db_obj:
                if db_obj[key] != val:
                    equal = False
                    break
        return equal

    def save_from_update_obj(self, update_obj):
        """Create or update object in DB"""
        ptb_obj = self.get_ptb_obj(update_obj)
        if ptb_obj:
            query = self.get_query(ptb_obj)
            db_obj = self.find_one(query)
            if db_obj:
                if not self.is_equal(ptb_obj, db_obj):
                    self.update_one(query, ptb_obj.to_dict())
            else:
                self.create(ptb_obj.to_dict())


class DefaultUpdateModel(ApiTypeModel):
    api_type = UpdateApiType

    name = api_type.name
    fields = api_type.fields

    def save_from_update_obj(self, update_obj):
        data = update_obj.to_dict()
        return self.create(data)

    def get_ptb_obj(self, update_obj):
        return update_obj

    def get_query(self, ptb_obj):
        query = {self.api_type.id_field: ptb_obj.update_id}
        return query


class DefaultMessageModel(ApiTypeModel):
    api_type = MessageApiType

    name = api_type.name
    fields = api_type.fields

    def get_ptb_obj(self, update_obj):
        ptb_obj = update_obj.effective_message
        return ptb_obj

    def get_query(self, ptb_obj):
        query = {self.api_type.id_field: ptb_obj.message_id, 'chat.id': ptb_obj.chat.id}
        return query


class DefaultUserModel(ApiTypeModel):
    api_type = UserApiType

    name = api_type.name
    fields = api_type.fields

    def get_ptb_obj(self, update_obj):
        ptb_obj = update_obj.effective_user
        return ptb_obj

    def get_query(self, ptb_obj):
        query = {self.api_type.id_field: ptb_obj.id}
        return query


class DefaultChatModel(ApiTypeModel):
    api_type = ChatApiType

    name = api_type.name
    fields = api_type.fields

    def get_ptb_obj(self, update_obj):
        ptb_obj = update_obj.effective_chat
        return ptb_obj

    def get_query(self, ptb_obj):
        query = {self.api_type.id_field: ptb_obj.id}
        return query


def get_model_classes():
    model_classes = [
        getattr(settings, k) for k in dir(settings)
        if k.endswith('_model_class')
    ]
    return model_classes


def mongo_get_first(cursor):
    """Return first item in the cursor"""
    return [item for item in cursor.limit(1)][0]


def mongo_get_last(cursor):
    """Return last item in the cursor"""
    return [item for item in cursor][-1]



import logging


tg_api_token = ''

db_name = ''
db_name_test = ''

db_host = 'localhost'
db_port = 27017

storage_class = 'meetg.storage.MongoStorage'
update_model_class = 'meetg.storage.DefaultUpdateModel'
message_model_class = 'meetg.storage.DefaultMessageModel'
user_model_class = 'meetg.storage.DefaultUserModel'
chat_model_class = 'meetg.storage.DefaultChatModel'

bot_class = 'bot.MyBot'

api_attempts = 5
network_error_wait = 2

log_path = 'log.txt'
log_level = logging.INFO

stats_to = ()



import datetime, time
from collections import defaultdict

import telegram
import pytz
from telegram.ext import Handler, Updater

import settings
from meetg.api_methods import api_method_classes
from meetg.loging import get_logger
from meetg.stats import (
    DateCache, get_job_reports, get_update_reports, _SaveTimeJobQueueWrapper, service_cache,
)
from meetg.storage import ApiTypeModel, get_model_classes
from meetg.testing import UpdaterMock
from meetg.factories import MessageUpdateFactory
from meetg.utils import (
    get_current_unixtime, get_unixtime_before_now, get_update_type, import_string,
)


logger = get_logger()


class BaseBot:
    """Common Telegram bot logic"""

    def __init__(self, mock=False):
        self._is_mock = mock
        self._init_updater()
        self._init_models(test=mock)
        self._init_handlers()
        self._init_jobs()
        self.last_method = None

    def _init_updater(self):
        """Init PTB updater"""
        updater_class = UpdaterMock if self._is_mock else Updater
        self.updater = updater_class(settings.tg_api_token, use_context=True)
        self._tgbot = self.updater.bot
        self.username = self.updater.bot.get_me().username

    def _init_handlers(self):
        service_handler = _ServiceHandler(self._models)
        self._handlers = (service_handler,) + self.init_handlers()
        if not self._is_mock:
            for handler in self._handlers:
                self.updater.dispatcher.add_handler(handler)

    def init_handlers(self):
        """Intended to be redefined in your bot class"""
        logger.warning('No handlers found')
        return ()

    def init_jobs(self, job_queue):
        """Intended to be redefined in your bot class"""
        pass

    def _init_jobs(self):
        """Set default jobs to self.updater.job_queue before self.init_jobs()"""
        self._job_queue_wrapper = _SaveTimeJobQueueWrapper(self.updater.job_queue)
        stats_dt = datetime.time(tzinfo=pytz.timezone('UTC'))  # 00:00 UTC
        self._job_queue_wrapper.run_daily(self._job_report_stats, stats_dt)
        self.init_jobs(self._job_queue_wrapper)

    def _init_models(self, test=False):
        """Read model classes from settings, import and add them to self"""
        self._models = []
        for model_class in get_model_classes():
            model = import_string(model_class)(test=test)
            setattr(self, f'{model.name_lower}_model', model)
            self._models.append(model)

    def _mock_process_update(self, update_obj):
        """Simulation of telegram.ext.dispatcher.Dispatcher.process_update()"""
        for handler in self._handlers:
            check = handler.check_update(update_obj)
            if check not in (None, False):
                return handler.callback(update_obj, None)

    def _job_report_stats(self, context=None):
        """Report bots stats daily"""
        job_reports = get_job_reports()
        update_reports = get_update_reports()
        model_reports = [m.get_day_report() for m in self._models]

        prefix = f'@{self.username} for the last 24 hours:'
        lines = [prefix] + update_reports + model_reports + job_reports
        report = '\n- '.join(lines)

        logger.info(report)
        if settings.stats_to:
            self.send_messages(settings.stats_to, report)

    def run(self):
        self.updater.start_polling()
        logger.info('@%s started', self.username)
        self.updater.idle()

    def send_messages(self, chat_ids, text, reply_to=None, markup=None, html=None, preview=False):
        """Shortcut to replace multiple send_message API calls"""
        for chat_id in chat_ids:
            self.send_message(
                chat_id, text, reply_to=reply_to, markup=markup, html=html, preview=preview,
            )
        logger.info('Message broadcasted: %s', repr(text[:79]))

    def receive_message(self, text='', **kwargs):
        """
        Simulates receiving Update with 'message' by the bot in tests
        """
        factory = MessageUpdateFactory(self, 'message')
        update_obj = factory.create(text=text, **kwargs)
        return self._mock_process_update(update_obj)

    def receive_edited_message(self, text, chat_id, message_id, **kwargs):
        """
        Simulates receiving Update with 'edited_message' by the bot in tests
        """
        factory = MessageUpdateFactory(self, 'edited_message')
        update_obj = factory.create(text=text, chat__id=chat_id, message_id=message_id, **kwargs)
        return self._mock_process_update(update_obj)

    def __getattr__(self, attrname):
        """
        Find API method class by the name, instantiate it,
        remember in self.last_method and return generated method
        with its easy_call() result inside
        """
        method_class = api_method_classes.get(attrname)
        if method_class:
            method_obj = method_class(self._tgbot, self._is_mock)
            self.last_method = method_obj

            def _internal_call(*args, **kwargs):
                return method_obj.easy_call(*args, **kwargs)

            return _internal_call

        else:
            raise NameError(f'API method {attrname} not found')


class _ServiceHandler(Handler):
    """
    Fake handler which handles no updates,
    but saves info from each Update for update-related models,
    if they are enabled, and count stats
    """
    def __init__(self, models):
        super().__init__(lambda: None)
        self.gather_models(models)

    def gather_models(self, models):
        """Gather models related to Bot API, to save them later"""
        self.models = []
        for model in models:
            if isinstance(model, ApiTypeModel) and model.fields:
                self.models.append(model)

    def check_update(self, update_obj):
        """The method triggers by PTB on each received update"""
        print(update_obj)
        self.save(update_obj)
        self.count(update_obj)

    def count(self, update_obj):
        """Count stats for a later report"""
        update_type = get_update_type(update_obj)
        service_cache['stats']['update'].init(DateCache)
        service_cache['stats']['update'][update_type].add()

    def save(self, update_obj):
        """Save all the fields specified in enabled models"""
        for model in self.models:
            model.save_from_update_obj(update_obj)



import os
import sys
import unittest

import settings
from meetg.utils import import_string


KNOWN_ARGS = ('run', 'test')


def run_bot(bot_path):
    Bot = import_string(bot_path)
    Bot().run()


def run_tests(import_pathes, src_path):
    """Run tests by import_pathes, or, if not provided, run all the tests"""
    loader = unittest.loader.TestLoader()
    if import_pathes:
        suite = loader.loadTestsFromNames(import_pathes)
    else:
        suite = loader.discover(src_path)

    result = unittest.runner.TextTestRunner().run(suite)


def exec_args(argv, src_path):
    if len(argv) > 1 and argv[1] in KNOWN_ARGS:
        if argv[1] == 'run':
            run_bot(settings.bot_class)
        if argv[1] == 'test':
            run_tests(argv[2:], src_path)
    else:
        print('Available commands:', ', '.join(KNOWN_ARGS))
